# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path

from aiconsole.core.gpt.consts import GPTMode
import litellm


# FIXME: Move it to a more appropriate place
litellm.set_verbose = False
MAX_BUDGET = None

if MAX_BUDGET:
    litellm.max_budget = MAX_BUDGET

AICONSOLE_PATH = Path(__file__).parent

ORIGINS: list[str] = [
    "http://localhost:3000",
    "http://localhost:3001",
    "http://127.0.0.1:3000",
    "http://127.0.0.1:3001",
]

LOG_FORMAT: str = "{asctime} {name} [{levelname}] {message}"
LOG_STYLE: str = "{"
LOG_LEVEL: str = "DEBUG"
LOG_HANDLERS: list[str] = ["console"]


HISTORY_LIMIT: int = 1000
COMMANDS_HISTORY_JSON: str = "command_history.json"

DEFAULT_MODE: str = GPTMode.SPEED.value
FUNCTION_CALL_OUTPUT_LIMIT: int = 2000


DIRECTOR_MIN_TOKENS: int = 250
DIRECTOR_PREFERRED_TOKENS: int = 1000

MAX_RECENT_PROJECTS = 8

log_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "console": {
            "()": "logging.Formatter",
            "fmt": LOG_FORMAT,
            "style": LOG_STYLE,
        }
    },
    "handlers": {
        "console": {
            "formatter": "console",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
        "sinkhole": {"class": "logging.NullHandler"},
    },
    "loggers": {
        "aiconsole": {
            "handlers": LOG_HANDLERS,
            "level": LOG_LEVEL,
            "propagate": False,
        },
        "uvicorn": {
            "handlers": LOG_HANDLERS,
            "level": "INFO",
        },
    },
}



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from aiconsole.core.assets.asset import AssetLocation
from fastapi import APIRouter
from fastapi.responses import JSONResponse
from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.chat.types import Chat
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.assets.materials.content_evaluation_context import ContentEvaluationContext
from aiconsole.core.assets.materials.material import Material

router = APIRouter()


@router.post("/preview")
async def materials_preview(material: Material):
    content_context = ContentEvaluationContext(
        chat=Chat(id="chat", name="", last_modified=datetime.now(), title_edited=False, message_groups=[]),
        agent=Agent(
            id="user",
            name="You",
            usage="",
            usage_examples=[],
            system="",
            defined_in=AssetLocation.AICONSOLE_CORE,
            gpt_mode=GPTMode.QUALITY,
            override=False,
        ),
        gpt_mode=GPTMode.SPEED,
        relevant_materials=[],
    )

    rendered_material = await material.render(content_context)

    return JSONResponse(rendered_material.model_dump())



import asyncio
from pathlib import Path
from typing import Any, Dict

from aiconsole.core.settings.project_settings import Settings


def set_code_autorun(autorun: bool) -> None:
    settings = Settings(Path("."))
    asyncio.run(settings.reload())
    settings.set_code_autorun(autorun)
    settings.stop()


def get_settings() -> Dict[str, Any]:
    settings = Settings()
    asyncio.run(settings.reload())
    settings.stop()
    return settings.model_dump()



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    
class NoOpenAPIKeyException(Exception):
    def __str__(self):
        return "No API key provided or is invalid. " \
               "You must set the environment variable OPENAI_API_KEY=<API-KEY>. " \
               "Generate API keys in the OpenAI web interface. " \
               "See https://platform.openai.com/account/api-keys for details."



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from aiconsole.core.assets.asset import AssetType
from aiconsole.core.project.paths import get_project_assets_directory
from aiconsole.core.settings.project_settings import get_aiconsole_settings


async def move_asset_in_fs(asset_type: AssetType, old_id: str, new_id: str) -> None:
    old_file_path = get_project_assets_directory(asset_type) / f"{old_id}.toml"
    new_file_path = get_project_assets_directory(asset_type) / f"{new_id}.toml"

    # Check if the old file exists
    if not old_file_path.exists():
        raise FileNotFoundError(f"'{old_id}' does not exist.")

    # Check if the new file already exists
    if new_file_path.exists():
        raise FileExistsError(f"'{new_id}' already exists.")

    # Move (rename) the file
    old_file_path.rename(new_file_path)
    get_aiconsole_settings().rename_asset(asset_type, old_id, new_id)



import os
from pathlib import Path
from typing import Dict

from aiconsole.api.websockets.outgoing_messages import ErrorWSMessage
from aiconsole.core.assets.asset import Asset, AssetLocation, AssetType
from aiconsole.core.assets.fs.load_asset_from_fs import load_asset_from_fs
from aiconsole.core.project.paths import get_core_assets_directory, get_project_assets_directory
from aiconsole.utils.list_files_in_file_system import list_files_in_file_system


async def load_all_assets(asset_type: AssetType) -> Dict[str, list[Asset]]:
    _assets = {}

    locations = [
        [AssetLocation.PROJECT_DIR, get_project_assets_directory(asset_type)],
        [AssetLocation.AICONSOLE_CORE, get_core_assets_directory(asset_type)],
    ]

    for [location, dir] in locations:
        ids = set(
            [
                os.path.splitext(os.path.basename(path))[0]
                for path in list_files_in_file_system(dir)
                if os.path.splitext(Path(path))[-1] == ".toml"
            ]
        )

        for id in ids:
            try:
                asset = await load_asset_from_fs(asset_type, id, location)
                if id not in _assets:
                    _assets[id] = []
                _assets[id].append(asset)
            except Exception as e:
                await ErrorWSMessage(
                    error=f"Invalid {asset_type} {id} {e}",
                ).send_to_all()
                continue

    return _assets



#
# MIT License
#
# Copyright (c) 2023 Killian Lucas
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#     ____                      ____      __                            __
#    / __ \____  ___  ____     /  _/___  / /____  _________  ________  / /____  _____
#   / / / / __ \/ _ \/ __ \    / // __ \/ __/ _ \/ ___/ __ \/ ___/ _ \/ __/ _ \/ ___/
#  / /_/ / /_/ /  __/ / / /  _/ // / / / /_/  __/ /  / /_/ / /  /  __/ /_/  __/ /
#  \____/ .___/\___/_/ /_/  /___/_/ /_/\__/\___/_/  / .___/_/   \___/\__/\___/_/
#      /_/                                         /_/
#
# This file has been taken from the wonderful project "open-interpreter" by Killian Lucas
# https://github.com/KillianLucas/open-interpreter
#

import sys
from typing import List

from aiconsole.core.assets.materials.material import Material
from ..subprocess_code_interpreter import SubprocessCodeInterpreter
import ast
import re
import logging

_log = logging.getLogger(__name__)


class Python(SubprocessCodeInterpreter):
    file_extension = "py"
    proper_name = "Python"

    def __init__(self):
        super().__init__()
        # the python command itself is provided by the virtual environment
        self.start_cmd = "python -i -q -u"

    def preprocess_code(self, code: str, materials: List[Material]):
        return preprocess_python(code, materials)

    def line_postprocessor(self, line):
        if re.match(r"^(\s*>>>\s*|\s*\.\.\.\s*)", line):
            return None
        return line

    def detect_end_of_execution(self, line):
        return "## end_of_execution ##" in line


def preprocess_python(code: str, materials: List[Material]):
    """
    Add active line markers
    Wrap in a try except
    Add end of execution marker
    """

    # Check for syntax errors in user's code
    try:
        ast.parse(code)
    except SyntaxError as e:
        # If there's a syntax error, return the error message directly
        newline = "\n"
        # msg_for_user = f"SyntaxError on line {e.lineno}, column {e.offset}: {e.msg} ({e.text})"

        msg_for_user = (
            f""
            f'File "{e.filename}", line {e.lineno}, column {e.offset}\n '
            f"  {(e.text or '').replace(newline, '')}\n"
            f"  {(e.offset or 0) * ' '}^\n"
            f"SyntaxError: {e.msg}\n"
        )

        return f"print(f'''{msg_for_user}''')\nprint('## end_of_execution ##')"

    api_materials = [material for material in materials if material.content_type == "api"]
    apis = [material.content_api for material in api_materials]

    parsed_code = ast.parse("\n\n\n".join(apis))
    parsed_code.body = [b for b in parsed_code.body if not isinstance(b, ast.Expr) or not isinstance(b.value, ast.Str)]
    apis = ast.unparse(parsed_code)

    newline = "\n"
    code = f"""
{apis}


import traceback
from aiconsole.dev.credentials import MissingCredentialException

try:
{newline.join(("    " + line) for line in code.split(newline) if line.strip())}
except MissingCredentialException as e:
    print(e)
except Exception:
    traceback.print_exc()


print("## end_of_execution ##")

""".strip()

    return code



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    

import json
import os
from typing import Dict, List

from aiconsole.core.project import project
from aiconsole.core.project.paths import get_credentials_directory


class MissingCredentialException(Exception):
    pass

def save_credential(module: str, credential: str, value: str):
    # Specify the path for the credential file
    file_path = get_credentials_directory() / (module + ".json")
    
    # Ensure the directory for the file exists
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    
    # Load the existing JSON object or create an empty one if the file doesn't exist
    try:
        with open(file_path, "r") as f:
            credentials = json.load(f)
    except FileNotFoundError:
        credentials = {}
    
    # Update the credential
    credentials[credential] = value
    
    # Save the updated credentials back to the JSON file
    with open(file_path, "w") as f:
        json.dump(credentials, f, indent=4)  # Use an indent of 4 for pretty printing

def load_credentials(module: str, credentials: List[str]) -> Dict[str, str]:
    result = {}

    for credential in credentials:
        # Try env first
        env_name = (module + "_" + credential).upper()
        if os.environ.get(env_name):
            result[credential] = os.environ.get(env_name) or ""
            continue
        
        # Try file
        try: 
            with open(get_credentials_directory() / (module + ".json"), "r") as f:
                value = str(json.loads(f.read())[credential])
                if value:
                    result[credential] = value
                    continue
        except FileNotFoundError:
            pass
        except KeyError:
            pass

    missing_credentials = [credential for credential in credentials if credential not in result]

    if missing_credentials:
        raise MissingCredentialException(f'''
This module ({module}) requires following credentials: {','.join(credentials)}.

Could not find credentials: {','.join(missing_credentials)}.

The user should provide those credentials, explain to them how to do it, and what they are.
Don't use 'getpass' or anything that reads from the console, ask them using text,
When they provide you with credentials, save the credentials for the user, using the following code:

from aiconsole.dev.credentials import save_credential
save_credential("{module}", "{missing_credentials[0]}", your_value)
    ...
        '''.strip())
    
    return result


from typing import Callable

from aiconsole.core.assets.asset import Asset, AssetLocation, AssetType
from aiconsole.core.project import project
from aiconsole.core.settings.project_settings import get_aiconsole_settings
from aiconsole.utils.capitalize_first import capitalize_first
from fastapi import HTTPException
from fastapi.responses import JSONResponse


async def asset_get(request, asset_type: AssetType, asset_id: str, new_asset: Callable[[], Asset]):
    location_param = request.query_params.get("location", None)
    location = AssetLocation(location_param) if location_param else None

    if asset_id == "new":
        asset = new_asset()
        asset.defined_in = AssetLocation.PROJECT_DIR
        asset.override = False
        return JSONResponse(asset.model_dump())
    else:
        settings = get_aiconsole_settings()

        if asset_type == AssetType.AGENT:
            assets = project.get_project_agents()
        elif asset_type == AssetType.MATERIAL:
            assets = project.get_project_materials()
        else:
            raise ValueError(f"Invalid asset type: {asset_type}")

        agent = assets.get_asset(asset_id, location)

        if not agent:
            raise HTTPException(status_code=404, detail=f"{capitalize_first(asset_type)} not found")

        # capitalize first letter

        return JSONResponse(
            {
                **agent.model_dump(),
                "status": settings.get_asset_status(asset_type, agent.id),
            }
        )



from aiconsole.core.assets.asset import AssetType
from aiconsole.core.project.paths import get_project_assets_directory


def project_asset_exists_fs(asset_type: AssetType, asset_id: str) -> bool:
    return (get_project_assets_directory(asset_type) / f"{asset_id}.toml").exists()



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import List, Optional, cast

from aiconsole.consts import DIRECTOR_MIN_TOKENS, DIRECTOR_PREFERRED_TOKENS
from aiconsole.core.analysis.agents_to_choose_from import agents_to_choose_from
from aiconsole.core.analysis.create_plan_class import create_plan_class
from aiconsole.core.assets.agents.agent import Agent
from aiconsole.core.assets.asset import AssetLocation, AssetStatus
from aiconsole.core.assets.materials.material import Material
from aiconsole.core.chat.chat_outgoing_messages import SequenceStage, UpdateAnalysisWSMessage
from aiconsole.core.chat.types import Chat
from aiconsole.core.gpt.consts import GPTMode
from aiconsole.core.gpt.gpt_executor import GPTExecutor
from aiconsole.core.gpt.request import GPTRequest, ToolDefinition, ToolFunctionDefinition
from aiconsole.core.gpt.types import EnforcedFunctionCall, EnforcedFunctionCallFuncSpec, GPTRequestTextMessage
from aiconsole.core.project import project
from aiconsole.utils.convert_messages import convert_messages

_log = logging.getLogger(__name__)


def pick_agent(arguments, chat: Chat, available_agents: list[Agent]) -> Agent:
    # Try support first
    default_agent = next((agent for agent in available_agents if agent.id == "assistant"), None)

    # Pick any if not available
    if not default_agent:
        default_agent = available_agents[0]

    is_users_turn = arguments.is_users_turn

    if is_users_turn:
        picked_agent = Agent(
            id="user",
            name="User",
            usage="When a human user needs to respond",
            usage_examples=[],
            system="",
            defined_in=AssetLocation.AICONSOLE_CORE,
            gpt_mode=GPTMode.QUALITY,
            override=False,
        )
    else:
        picked_agent = next(
            (agent for agent in available_agents if agent.id == arguments.agent_id),
            None,
        )

    _log.debug(f"Chosen agent: {picked_agent}")

    if not picked_agent:
        picked_agent = default_agent

    # If it turns out that the user must respond to him self, have the assistant drive the conversation
    if is_users_turn and chat.message_groups and chat.message_groups[-1].role == "user":
        picked_agent = default_agent

    return picked_agent


def _get_relevant_materials(relevant_material_ids: List[str]) -> List[Material]:
    # Maximum of 5 materials
    relevant_materials = [
        cast(Material, k)
        for k in project.get_project_materials().assets_with_status(AssetStatus.ENABLED)
        if k.id in relevant_material_ids
    ][:5]

    relevant_materials += cast(list[Material], project.get_project_materials().assets_with_status(AssetStatus.FORCED))

    return relevant_materials


async def gpt_analysis_function_step(
    chat: Chat,
    request_id: str,
    gpt_mode: GPTMode,
    initial_system_prompt: str,
    last_system_prompt: str,
    force_call: bool,
):
    gpt_executor = GPTExecutor()

    # Pick from forced or enabled agents if no agent is forced
    possible_agent_choices = agents_to_choose_from()

    if len(possible_agent_choices) == 0:
        raise ValueError("No active agents")

    plan_class = create_plan_class(
        [
            Agent(
                id="user",
                name="User",
                usage="When a human user needs to respond",
                usage_examples=[],
                system="",
                defined_in=AssetLocation.AICONSOLE_CORE,
                override=False,
            ),
            *possible_agent_choices,
        ]
    )

    request = GPTRequest(
        system_message=initial_system_prompt,
        gpt_mode=gpt_mode,
        messages=[*convert_messages(chat), GPTRequestTextMessage(role="system", content=last_system_prompt)],
        tools=[ToolDefinition(type="function", function=ToolFunctionDefinition(**plan_class.openai_schema))],
        presence_penalty=2,
        min_tokens=DIRECTOR_MIN_TOKENS,
        preferred_tokens=DIRECTOR_PREFERRED_TOKENS,
    )

    if force_call:
        request.tool_choice = EnforcedFunctionCall(
            type="function", function=EnforcedFunctionCallFuncSpec(name=plan_class.__name__)
        )

    await UpdateAnalysisWSMessage(request_id=request_id, stage=SequenceStage.START).send_to_chat(chat.id)

    try:
        async for chunk in gpt_executor.execute(request):
            if len(gpt_executor.partial_response.choices) > 0:
                tool_calls = gpt_executor.partial_response.choices[0].message.tool_calls
                for tool_call in tool_calls:
                    function_call = tool_call.function
                    arguments = function_call.arguments
                    if not isinstance(arguments, str):
                        await UpdateAnalysisWSMessage(
                            stage=SequenceStage.MIDDLE,
                            request_id=request_id,
                            agent_id=arguments.get("agent_id", None),
                            relevant_material_ids=arguments.get("relevant_material_ids", None),
                            next_step=arguments.get("next_step", None),
                            thinking_process=arguments.get("thinking_process", None),
                        ).send_to_chat(chat.id)
                if not tool_calls:
                    await UpdateAnalysisWSMessage(
                        stage=SequenceStage.MIDDLE,
                        request_id=request_id,
                        agent_id=None,
                        relevant_material_ids=None,
                        next_step=None,
                        thinking_process=gpt_executor.partial_response.choices[0].message.content,
                    ).send_to_chat(chat.id)
            else:
                _log.warning("No choices in partial response")
                _log.warning(chunk)

        result = gpt_executor.response.choices[0].message

        if len(result.tool_calls) == 0:
            await UpdateAnalysisWSMessage(
                request_id=request_id,
                next_step=result.content or "",
                stage=SequenceStage.MIDDLE,
            ).send_to_chat(chat.id)
            return

        if len(result.tool_calls) > 1:
            raise ValueError(f"Expected one tool call, got {len(result.tool_calls)}")

        arguments = result.tool_calls[0].function.arguments

        if isinstance(arguments, str):
            raise ValueError(f"Could not parse arguments from the text: {arguments}")

        arguments = plan_class(**arguments)

        picked_agent = pick_agent(arguments, chat, possible_agent_choices)

        relevant_materials = _get_relevant_materials(arguments.relevant_material_ids)

        await UpdateAnalysisWSMessage(
            request_id=request_id,
            next_step=arguments.next_step,
            agent_id=picked_agent.id,
            relevant_material_ids=[material.id for material in relevant_materials],
            stage=SequenceStage.MIDDLE,
        ).send_to_chat(chat.id)
    finally:
        await UpdateAnalysisWSMessage(
            request_id=request_id,
            stage=SequenceStage.END,
        ).send_to_chat(chat.id)



#
# MIT License
#
# Copyright (c) 2023 Killian Lucas
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#     ____                      ____      __                            __           
#    / __ \____  ___  ____     /  _/___  / /____  _________  ________  / /____  _____
#   / / / / __ \/ _ \/ __ \    / // __ \/ __/ _ \/ ___/ __ \/ ___/ _ \/ __/ _ \/ ___/
#  / /_/ / /_/ /  __/ / / /  _/ // / / / /_/  __/ /  / /_/ / /  /  __/ /_/  __/ /    
#  \____/ .___/\___/_/ /_/  /___/_/ /_/\__/\___/_/  / .___/_/   \___/\__/\___/_/     
#      /_/                                         /_/                               
#
# This file has been taken from the wonderful project "open-interpreter" by Killian Lucas
# https://github.com/KillianLucas/open-interpreter
#

from typing import Dict, Literal, Type, Union
from aiconsole.core.code_running.code_interpreters.base_code_interpreter import BaseCodeInterpreter
from .languages.python import Python
from .languages.shell import Shell
from .languages.applescript import AppleScript

LanguageStr = Union[Literal["python"], Literal["shell"], Literal["applescript"]]

language_map: Dict[LanguageStr, Type[BaseCodeInterpreter]] = {
    "python": Python,
    "shell": Shell,
    "applescript": AppleScript,
}


# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import cast
from aiconsole.api.utils.asset_save import asset_patch, asset_post
from aiconsole.core.assets.get_material_content_name import get_material_content_name
from aiconsole.api.utils.asset_exists import asset_exists
from aiconsole.api.utils.asset_get import asset_get
from aiconsole.api.utils.asset_status_change import asset_status_change
from aiconsole.api.utils.status_change_post_body import StatusChangePostBody
from aiconsole.core.assets.asset import AssetLocation, AssetStatus, AssetType
from aiconsole.core.assets.materials.material import Material, MaterialContentType, MaterialWithStatus
from aiconsole.core.project import project
from fastapi import APIRouter, HTTPException, Request
from fastapi.responses import JSONResponse

router = APIRouter()


@router.get("/{material_id}")
async def material_get(request: Request, material_id: str):
    type = cast(MaterialContentType, request.query_params.get("type", ""))

    return await asset_get(
        request,
        AssetType.MATERIAL,
        material_id,
        lambda: MaterialWithStatus(
            id="",
            name="New " + get_material_content_name(type),
            usage="",
            usage_examples=[],
            status=AssetStatus.ENABLED,
            defined_in=AssetLocation.PROJECT_DIR,
            override=False,
        ),
    )


@router.patch("/{asset_id}")
async def agent_patch(asset_id: str, material: Material):
    return await asset_patch(AssetType.MATERIAL, material, asset_id)


@router.post("/{asset_id}")
async def agent_post(asset_id: str, material: Material):
    return await asset_post(AssetType.MATERIAL, material, asset_id)


@router.post("/{material_id}/status-change")
async def material_status_change(material_id: str, body: StatusChangePostBody):
    return await asset_status_change(AssetType.MATERIAL, material_id, body)


@router.delete("/{material_id}")
async def delete_material(material_id: str):
    try:
        await project.get_project_materials().delete_asset(material_id)
        return JSONResponse({"status": "ok"})
    except KeyError:
        raise HTTPException(status_code=404, detail="Material not found")


@router.get("/{asset_id}/exists")
async def material_exists(request: Request, asset_id: str):
    return await asset_exists(AssetType.MATERIAL, request, asset_id)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
    

import asyncio
from functools import wraps
import inspect
import logging
import threading
import time
from fastapi import Request
from fastapi.concurrency import run_until_first_complete
from fastapi.responses import StreamingResponse

_log = logging.getLogger(__name__)

def disconnect_aware_generator(original_gen, request):
    client_disconnected = threading.Event()

    # Thread target function to check for client disconnection
    def check_disconnection():
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            while not client_disconnected.is_set():
                if loop.run_until_complete(request.is_disconnected()):
                    client_disconnected.set()
                time.sleep(0.5)  # Check every 0.5 seconds
        except Exception:
            pass
        finally:
            loop.close()

    async def async_gen():
        if hasattr(original_gen, '__aiter__'):
            # Asynchronous generator
            async for chunk in original_gen:
                if client_disconnected.is_set():
                    raise asyncio.CancelledError("Client disconnected")
                yield chunk
        else:
            # Synchronous generator
            disconnection_thread = threading.Thread(target=check_disconnection)
            disconnection_thread.start()
            try:
                for chunk in original_gen:
                    if client_disconnected.is_set():
                        raise asyncio.CancelledError("Client disconnected")
                    yield chunk
            finally:
                client_disconnected.set()
                disconnection_thread.join()

    return async_gen()


async def raise_if_disconnected(request: Request):
    while True:
        if await request.is_disconnected():
            raise asyncio.CancelledError("Client disconnected")
        await asyncio.sleep(0.1)  # Check every 0.1 seconds

def cancelable_endpoint(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        request = kwargs.get('request')
        if not request:
            raise ValueError("The endpoint must have a 'request' argument")
        
        response = None

        async def func_wrapper():
            nonlocal response
            response = await func(*args, **kwargs)

        await run_until_first_complete(
            (func_wrapper, {}),
            (raise_if_disconnected, {"request": request})
        )

        if isinstance(response, StreamingResponse):
            response.body_iterator = disconnect_aware_generator(response.body_iterator, request)

        return response

    return wrapper


# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
from typing import Dict, List, Optional
from aiconsole.core.assets.asset import Asset, AssetLocation, AssetStatus, AssetType
from aiconsole.core.assets.fs.move_asset_in_fs import move_asset_in_fs
from aiconsole.core.assets.fs.project_asset_exists_fs import project_asset_exists_fs
from aiconsole.core.assets.fs.delete_asset_from_fs import delete_asset_from_fs
from aiconsole.core.assets.fs.save_asset_to_fs import save_asset_to_fs
from aiconsole.core.assets.load_all_assets import load_all_assets

import watchdog.events
import watchdog.observers
from aiconsole.core.project.paths import get_project_assets_directory
from aiconsole.core.settings.project_settings import get_aiconsole_settings
from aiconsole.utils.BatchingWatchDogHandler import BatchingWatchDogHandler
from aiconsole.api.websockets.outgoing_messages import AssetsUpdatedWSMessage

_log = logging.getLogger(__name__)


class Assets:
    # _assets have lists, where the first element is the one overriding the others (currently there can be only one overriden element)
    _assets: Dict[str, list[Asset]]

    def __init__(self, asset_type: AssetType):
        self._suppress_notification_until = None
        self.asset_type = asset_type
        self._assets = {}

        self.observer = watchdog.observers.Observer()

        get_project_assets_directory(asset_type).mkdir(parents=True, exist_ok=True)
        self.observer.schedule(
            BatchingWatchDogHandler(self.reload), get_project_assets_directory(asset_type), recursive=True
        )
        self.observer.start()

    def stop(self):
        self.observer.stop()

    def all_assets(self) -> List[Asset]:
        """
        Return all loaded assets.
        """
        return list(assets[0] for assets in self._assets.values())

    def assets_with_status(self, status: AssetStatus) -> List[Asset]:
        """
        Return all loaded assets with a specific status.
        """
        settings = get_aiconsole_settings()
        return [
            assets[0]
            for assets in self._assets.values()
            if settings.get_asset_status(self.asset_type, assets[0].id) in [status]
        ]

    async def save_asset(self, asset: Asset, old_asset_id: str, create: bool):
        if asset.defined_in != AssetLocation.PROJECT_DIR:
            raise Exception("Cannot save asset not defined in project.")

        exists_in_project = project_asset_exists_fs(self.asset_type, asset.id)
        old_exists = project_asset_exists_fs(self.asset_type, old_asset_id)

        if create and exists_in_project:
            raise Exception(f"Asset {asset.id} already exists.")

        if not create and not exists_in_project:
            raise Exception(f"Asset {asset.id} does not exist.")

        rename = False

        if create and old_asset_id and not exists_in_project and old_exists:
            await move_asset_in_fs(asset.type, old_asset_id, asset.id)
            rename = True

        new_asset = await save_asset_to_fs(asset)

        if asset.id not in self._assets:
            self._assets[asset.id] = []

        # integrity checks and deleting old assets from structure
        if not create:
            if not self._assets[asset.id] or self._assets[asset.id][0].defined_in != AssetLocation.PROJECT_DIR:
                raise Exception(f"Asset {asset.id} cannot be edited")
            self._assets[asset.id].pop(0)
        else:
            if self._assets[asset.id] and self._assets[asset.id][0].defined_in == AssetLocation.PROJECT_DIR:
                raise Exception(f"Asset {asset.id} already exists")

        self._assets[asset.id].insert(0, new_asset)

        self._suppress_notification()

        return rename

    async def delete_asset(self, asset_id):
        self._assets[asset_id].pop(0)

        if len(self._assets[asset_id]) == 0:
            del self._assets[asset_id]

        delete_asset_from_fs(self.asset_type, asset_id)

        self._suppress_notification()

    def _suppress_notification(self):
        self._suppress_notification_until = datetime.datetime.now() + datetime.timedelta(seconds=10)

    def get_asset(self, id, location: AssetLocation | None = None):
        """
        Get a specific asset.
        """
        if id not in self._assets or len(self._assets[id]) == 0:
            return None

        for asset in self._assets[id]:
            if location == None or asset.defined_in == location:
                return asset

        return None

    async def reload(self, initial: bool = False):
        _log.info(f"Reloading {self.asset_type}s ...")

        self._assets = await load_all_assets(self.asset_type)

        await AssetsUpdatedWSMessage(
            initial=(
                initial
                or not (
                    not self._suppress_notification_until
                    or self._suppress_notification_until < datetime.datetime.now()
                )
            ),
            asset_type=self.asset_type,
            count=len(self._assets),
        ).send_to_all()



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from aiconsole.core.project import project

from aiconsole.api.websockets import connection_manager
from aiconsole.api.websockets.handle_incoming_message import handle_incoming_message

router = APIRouter()

_log = logging.getLogger(__name__)


@router.websocket("/ws")
async def websocket_endpoint(websocket: WebSocket):
    connection = await connection_manager.connect(websocket)
    await project.send_project_init(connection)

    try:
        while True:
            json_data = await websocket.receive_json()
            await handle_incoming_message(connection, json_data)
    except WebSocketDisconnect:
        connection_manager.disconnect(connection)



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from aiconsole.core.assets.asset import AssetType
from aiconsole.core.project.paths import get_project_assets_directory
from send2trash import send2trash


def delete_asset_from_fs(asset_type: AssetType, id):
    """
    Delete a specific agent.
    """

    # check if the file exists in project directory
    asset_file_path = get_project_assets_directory(asset_type) / f"{id}.toml"
    if asset_file_path.exists():
        send2trash(asset_file_path)
        return

    raise KeyError(f"{asset_type} with ID {id} not found")



# The AIConsole Project
#
# Copyright 2023 10Clouds
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging
from typing import Any, Dict, List, Literal, Optional, Union
from pydantic import BaseModel

import tiktoken

from aiconsole.core.gpt.consts import MODEL_DATA, GPTMode, GPTModel
from aiconsole.core.gpt.types import EnforcedFunctionCall, GPTRequestMessage, GPTRequestTextMessage
from aiconsole.core.gpt.token_error import TokenError

_log = logging.getLogger(__name__)

EXTRA_BUFFER_FOR_ENCODING_OVERHEAD = 50


class ToolFunctionParameters(BaseModel):
    type: "object"
    properties: Dict[str, Any]
    required: list[str]


class ToolFunctionDefinition(BaseModel):
    name: str
    description: str
    parameters: ToolFunctionParameters


class ToolDefinition(BaseModel):
    type: Literal["function"]
    function: ToolFunctionDefinition


class GPTRequest:
    def __init__(
        self,
        system_message: str,
        messages: List[GPTRequestMessage],
        gpt_mode: GPTMode,
        tools: List[ToolDefinition] = [],
        tool_choice: Union[Literal["none"], Literal["auto"], EnforcedFunctionCall, None] = None,
        temperature: float = 1,
        presence_penalty: float = 0,
        min_tokens: int = 0,
        preferred_tokens: int = 0,
    ):
        self.system_message = system_message
        self.messages = messages
        self.tools = tools or []
        self.tool_choice = tool_choice
        self.temperature = temperature
        self.gpt_mode = gpt_mode
        self.presence_penalty = presence_penalty
        self.max_tokens = 0

        # Checks if the given prompt can fit within a specified range of token lengths for the specified AI model.

        used_tokens = self.count_tokens() + EXTRA_BUFFER_FOR_ENCODING_OVERHEAD
        available_tokens = self.model_max_tokens - used_tokens

        if available_tokens < min_tokens:
            _log.error(
                f"Not enough tokens to perform the modification. Used tokens: {used_tokens},"
                f" available tokens: {available_tokens},"
                f" requested tokens: {self.max_tokens}"
            )

            raise TokenError(
                f"Exceeded the token limit by {min_tokens - available_tokens}, delete/edit some messages or reorganise materials."
            )

        self.max_tokens = min(available_tokens, preferred_tokens)

    def get_messages_dump(self):
        return [message.model_dump() for message in self.all_messages]

    @property
    def all_messages(self):
        if self.system_message:
            return [GPTRequestTextMessage(role="system", content=self.system_message), *self.messages]
        else:
            return self.messages

    @property
    def model(self):
        return self.get_model(self.gpt_mode)

    @property
    def model_data(self):
        return MODEL_DATA[self.model]

    @property
    def model_max_tokens(self):
        if self.gpt_mode == GPTMode.SPEED:
            return MODEL_DATA[GPTModel.GPT_35_TURBO_16k_0613].max_tokens

        return self.model_data.max_tokens

    def count_tokens(self):
        encoding = tiktoken.encoding_for_model(self.model_data.encoding)

        if self.tools:
            functions_tokens = len(encoding.encode(",".join(json.dumps(f.model_dump()) for f in self.tools)))
        else:
            functions_tokens = 0
        return self.count_messages_tokens(encoding) + functions_tokens

    def count_tokens_for_model(self, model):
        encoding = tiktoken.encoding_for_model(MODEL_DATA[model].encoding)
        return self.count_messages_tokens(encoding)

    def count_messages_tokens(self, encoding):
        messages_str = json.dumps(self.get_messages_dump())
        messages_tokens = len(encoding.encode(messages_str))

        return messages_tokens

    def count_tokens_output(self, message_content: str, message_function_call: Optional[Dict]):
        encoding = tiktoken.encoding_for_model(self.model_data.encoding)

        return len(encoding.encode(message_content)) + (
            len(encoding.encode(json.dumps(message_function_call))) if message_function_call else 0
        )

    def validate_request(self):
        """
        Checks if the prompt can be handled by the model
        """

        used_tokens = self.count_tokens()
        model_max_tokens = self.model_max_tokens

        if used_tokens - model_max_tokens >= self.max_tokens:
            _log.error(
                f"Not enough tokens to perform the modification. Used tokens: {used_tokens},"
                f" available tokens: {model_max_tokens},"
                f" requested tokens: {self.max_tokens}"
            )
            raise TokenError(
                f"Exceeded the token limit by {self.max_tokens - (used_tokens - model_max_tokens)}, delete/edit some messages or reorganise materials."
            )

    def get_model(self, mode: GPTMode) -> str:
        model = GPTModel.GPT_4_11106_PREVIEW
        # if mode == GPTMode.SPEED:
        #     model = GPTModel.GPT_35_TURBO_16k_0613
        #     used_tokens = self.count_tokens_for_model(model) + self.max_tokens
        #
        #     if used_tokens < MODEL_DATA[GPTModel.GPT_35_TURBO_0613].max_tokens * 0.9:
        #        model = GPTModel.GPT_35_TURBO_0613

        return model.value
